from .chart import Chart
import matplotlib.pyplot as plt
import plotly.express as px

class StackedAreaChart(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the StackedAreaChart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating StackedAreaChart visualization

        Returns:
            (string) date_label: date label  for axis-x
            (list) numerical_columns: numerical list
        """
        date_label = None
        numerical_column = None

        if self._is_date_column_exist(1):
            date_label = self._date_column[0]
            if self._is_numerical_column_exist(1):
                numerical_column = self._numerical_column
        
        return date_label, numerical_column          
  

    def plot(self):
        """
        Generate StackedAreaChart visualization
        """
        date_label, numerical_column  = self._check_requirements()

        if date_label is not None and numerical_column is not None:
            #set index by date label
            dataframe = self.dataframe.copy()
            dataframe = dataframe.set_index(date_label)
            #plot
            ax = dataframe.plot.area(stacked=True, figsize=(15,10))
            plt.show(block=True)




from .chart import Chart
import plotly.figure_factory as ff

class Table(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Table object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def plot(self):
        """
        Generate table visualization
        """
        fig = ff.create_table(self.dataframe)
        fig.show()    


from .chart import Chart
import plotly.express as px

class Histogram(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Histogram object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating Histogram visualization

        Returns:
            (string) numerical_label: label of numerical column
            (string) label_name: label name
        """
        numerical_label = None
        label_name = None

        if self._is_numerical_column_exist(1):
            numerical_label = self._numerical_column[0]
            if len(self._label_column) > 0:
                unique_list = [len(self.dataframe[name].unique()) for name in self._label_column]
                min_unique = min(unique_list)
                idx_min_unique = unique_list.index(min(unique_list))
                if min_unique <= 5 and min_unique < len(self.dataframe):
                    label_name=self._label_column[idx_min_unique]

        return numerical_label, label_name      

    def plot(self):
        """
        Generate Histogram visualization
        """
        numerical_label, label_name  = self._check_requirements()

        if numerical_label is not None:
            if label_name is not None:
                #plot
                fig = px.histogram(self.dataframe, x=numerical_label, color=label_name, marginal="rug", hover_data=self.dataframe.columns)
                fig.show()
            else:
                #plot
                fig = px.histogram(self.dataframe, x=numerical_label, marginal="rug", hover_data=self.dataframe.columns)
                fig.show()  




from .chart import Chart
from .bubblechart import BubbleChart
from .wordcloud import WordCloud
from .tree import Tree
from .graph import Graph
from .map import Map
from .table import Table
from .imagegrid import ImageGrid
from .dimensions import Dimensions
from .timeline import Timeline
from .scatterchart import ScatterChart
from .heatmap import HeatMap
from .histogram import Histogram
from .densityplot import DensityPlot
from .boxplot import BoxPlot
from .violinplot import ViolinPlot
from .donutchart import DonutChart
from .piechart import PieChart
from .sunburstchart import SunBurstChart
from .treemaps import TreeMap
from .barchart import BarChart
from .linechart import LineChart
from .areachart import AreaChart
from .stackedarechart import StackedAreaChart


from .chart import Chart
import matplotlib.pyplot as plt
import seaborn as sns

class HeatMap(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the HeatMap object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def plot(self):
        """
        Generate HeatMap visualization
        """

        if self._is_numerical_column_exist(3):
            #plot HeatMap
            plt.figure(figsize=(13,8))
            sns.heatmap(self.dataframe.corr(), annot = True)
            plt.show(block=True)


from .chart import Chart
import plotly.express as px
import datetime

class Timeline(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Timeline object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for Timeline visualization

        Returns:
            (list) date_column: label for axis-x
            (list) label_name: label for axis-y
        """
        date_column = None
        label_name = None

        if self._is_date_column_exist(1):
            date_column = self._date_column
            if len(self._label_column) == 0:
                if len(self._uri_column) > 0:
                    label_name = self._uri_column[0]
                elif hasattr(self.dataframe, 'coordinate'):
                    label_name = [name for name in self.dataframe.columns if name.startswith(tuple(['coordinate']))][0]
                else:
                    self._is_label_column_exist()
                    label_name = None
            else:
                label_name = self._label_column[0]
        
        return date_column, label_name


    def plot(self):
        """
        Generate Timeline visualization
        """
        date_column, label_name = self._check_requirements()

        if date_column is not None and label_name is not None:
            if len(date_column) >= 2:
                if self.dataframe[date_column[0]][0] > self.dataframe[date_column[1]][0]:
                    date_column[1],date_column[0] = date_column[0],date_column[1]
                fig = px.timeline(self.dataframe, x_start=date_column[0], x_end=date_column[1], 
                                y=label_name, color=label_name)
                fig.update_yaxes(autorange="reversed")
                fig.show()
            else:
                add_column = self.dataframe.copy()
                add_column['Y+1'] = [add_column[date_column[0]][i] + datetime.timedelta(days=365) for i in range (len(add_column))]
                fig = px.timeline(add_column, x_start=date_column[0], x_end='Y+1', 
                                    y=label_name, color=label_name, hover_data={'Y+1':False})
                fig.update_yaxes(autorange="reversed")
                fig.show()



from .chart import Chart
import matplotlib.pyplot as plt
import plotly.express as px
import seaborn as sns

class LineChart(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the LineChart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating LineChart visualization

        Returns:
            (string) date_label: date label  for axis-x
            (string) int_label: numerical label for axis-y
            (string) label_name: label for hue
        """
        date_label = None
        int_label = None
        label_name = None

        if self._is_date_column_exist(1):
            date_label = self._date_column[0]
            if self._is_numerical_column_exist(1):
                int_label = self._numerical_column[0]
                if len(self._label_column) > 0:
                    label_name = self._label_column[0]
        
        return date_label, int_label, label_name      

    def plot(self):
        """
        Generate LineChart visualization
        """
        date_label, numerical_label, label_name  = self._check_requirements()

        if date_label is not None and numerical_label is not None:
            if label_name is not None:
                plt.figure(figsize=(15,8))
                sns.lineplot(data=self.dataframe, x=date_label, y=numerical_label, hue=label_name)
                plt.show(block=True)
            else:
                fig = px.line(data_frame=self.dataframe, x=date_label, y=numerical_label)
                fig.show()




from .chart import Chart
import matplotlib.pyplot as plt
from imageio import imread
import time

class ImageGrid(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Image Grid visualization

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for Image Grid visualization

        Returns:
            (list) label_name: list of image label
        """
        label_name = None
        if hasattr(self.dataframe, 'picture'):
            if len(self._label_column) > 0:
                label_name = self._label_column[0]
            else:
                pass
        else:
            print(f"Data has no attribute 'picture'")
            print(self._add_candidate_info)
        
        return label_name

    def plot(self):
        """
        Generate Image Grid visualization
        """
        label_name = self._check_requirements()
        pic = [i for i in self.dataframe.picture]
        num_pic = len(pic)
        columns = 4
        width = 20
        height = max(20, int(num_pic/columns) * 20)

        if label_name is not None:
            item_label = [i for i in self.dataframe[label_name]]
            plt.figure(figsize=(20,20))
            for i, url in enumerate(pic):
                plt.subplot(int(num_pic / columns + 1), columns, i + 1)
                try:
                    image = imread(url)
                    plt.title(item_label[i])
                    plt.imshow(image) #, plt.xticks([]), plt.yticks([])
                    plt.axis('off')
                except:
                    time.sleep(5)
                    image = imread(url)
                    plt.title(item_label[i])
                    plt.imshow(image) #, plt.xticks([]), plt.yticks([])
                    plt.axis('off')  
        else:     
            plt.figure(figsize=(20,20))
            for i, url in enumerate(pic):
                plt.subplot(int(num_pic / columns + 1), columns, i + 1)
                try:
                    image = imread(url)
                    plt.imshow(image) #, plt.xticks([]), plt.yticks([])
                    plt.axis('off')
                except:
                    time.sleep(5)
                    image = imread(url)
                    plt.imshow(image) #, plt.xticks([]), plt.yticks([])
                    plt.axis('off')


import sys

from .utils import set_chart, set_dataframe
from .charts import Chart
from .chartdict import chartdict

class vizKG:
  """
  Instantiate vizKG object.
  
  Attributes:
      sparql_query (string): The SPARQL query to retrieve.
      sparql_service_url (string): The SPARQL endpoint URL.
      chart (string): Type of visualization
                     Options = {'Table', 'ImageGrid', 'Timeline' 'Graph' 
                                'Map', 'Tree','WordCloud', 'Dimensions',
                                'LineChart', 'BarChart', 'Histogram',
                                'DensityPlot', 'TreeMap' ,'SunBurstChart', 
                                'HeatMap' ,'PieChart', 'DonutChart',
                                'BoxPlot' ,'ViolinPlot', 'AreaChart',
                                'StackedAreaChart', 'ScatterChart', 'BubbleChart'}.
      **mode_keyword (boolean): Mode of mapping variable                           
      **figsize (float, float): Width, height in inches of plot
                
  """

  def __init__(self, sparql_query, sparql_service_url, chart=None, **kwargs):
      """
      Constructs all the necessary attributes for the vizKG object

      Parameters:
          sparql_query (string): The SPARQL query to retrieve.
          sparql_service_url (string): The SPARQL endpoint URL.
          chart (string): Type of visualization
          **mode_keyword (boolean): Mode of mapping variable 
          **figsize (float, float): Width, height in inches of plot

      """

      self.sparql_query = sparql_query
      self.sparql_service_url = sparql_service_url
      self.chart = set_chart(chart)
      self.kwargs = kwargs

      self.dataframe = set_dataframe(sparql_query, sparql_service_url)

  def plot(self):
      """
      Plot visualization with suitable corresponding chart

      """
      chart_list = chartdict.keys()
      chart = Chart(self.dataframe, self.kwargs)
      candidate_visualization = chart.candidate_form()
      figure = None
      if len(self.dataframe) != 0:
        if self.chart not in chart_list:
          print(f"No matching chart found instead use one of the available chart: {candidate_visualization}")
        else:
          try:
            if self.chart in chartdict:
              figure = chartdict[self.chart](self.dataframe, self.kwargs)
          finally:
            figure.plot()
      else:
        print("No matching records found")

sys.modules[__name__] = vizKG


from .chart import Chart
import plotly.express as px
import statistics


class PieChart(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the PieChart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating PieChart visualization

        Returns:
            (string) label_name: label name
            (list) numerical_column: list of numerical column
        """
        label_name = None
        numerical_column = None
        
        if self._is_numerical_column_exist(1):
            numerical_column = self._numerical_column
            if self._is_label_column_exist(1):
                if len(self._label_column) > 1:
                    axis_label, group_label = self._check_labels()
                    if group_label is not None:
                        label_name = group_label
                    else:
                        label_name = axis_label
                else:    
                    label_name = self._label_column[0]

        
        return label_name, numerical_column    

    def plot(self):
        """
        Generate PieChart visualization
        """
        label_name, numerical_column  = self._check_requirements()

        if label_name is not None and numerical_column is not None:
            values_label,hover_label = self._check_numerical_columns()
            if hover_label is not None:
                #plot
                fig = px.pie(self.dataframe, values=values_label, names=label_name,
                                hover_data=[hover_label])
                fig.show()
            else:
                fig = px.pie(self.dataframe, values=values_label, names=label_name)
                fig.show()                






from .chart import Chart
import seaborn as sns
import plotly.express as px

class BarChart(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the BarChart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating BarChart visualization

        Returns:
            (string) int_label: numerical label 
            (list) label_column: label column
        """
        int_label = None
        label_column = None

        if self._is_numerical_column_exist(1):
            int_label = self._numerical_column[0]
            if self._is_label_column_exist(1):
                label_column = self._label_column
        
        return int_label, label_column    

    def plot(self):
        """
        Generate BarChart visualization
        """
        numerical_label, label_column  = self._check_requirements()

        if label_column is not None and numerical_label is not None:
            axis_label,group_label,make_axis_label = None,None, None
            if len(label_column) >= 3:
                axis_label,group_label,make_axis_label = self._check_labels()
            else:
                axis_label,group_label = self._check_labels()
                
            orientation = self._check_orientation(axis_label,group_label)

            if make_axis_label is not None:
                axis_label = make_axis_label
            else:
                pass

            if group_label is not None:
                if orientation is not None:
                    fig = px.bar(self.dataframe, x=numerical_label, y=axis_label, color=group_label)
                    fig.show()
                else:
                    fig = px.bar(self.dataframe, x=axis_label, y=numerical_label, color=group_label)
                    fig.show()
            else:
                if orientation is not None:
                    fig = px.bar(self.dataframe, x=numerical_label, y=axis_label)
                    fig.show()
                else:
                    fig = px.bar(self.dataframe, x=axis_label, y=numerical_label)
                    fig.show() 



from .chart import Chart
import pandas as pd
import matplotlib.pyplot as plt
from wordcloud import STOPWORDS, WordCloud as wrdcld

class WordCloud(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the WordCloud object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)
        self._words = self._word_result()

    def _word_result(self):
        """
        Compile dataframe to one variable

        Parameters:
            (pandas.Dataframe) dataframe: The dataframe

        Returns:
            (string) words: The word result
        """
        #Merge into one column
        new_data = self.dataframe[self._label_column]
        new_data_flat = list(pd.Series(new_data.values.ravel('F')))

        #Merge into one variable
        words = " ".join([str(element) for element in new_data_flat])

        return words

    def plot(self):
        """
        Display WordCloud visualizations

        Parameters:
            (string) words: the visualized words
        """
        if self._is_label_column_exist(1):
            #initiate words
            words = self._word_result()
            #initiate wordcloud object
            stopwords = set(STOPWORDS) 
            wordcloud = wrdcld(
                            width = 800, height = 800, 
                            background_color ='white', 
                            stopwords = stopwords, 
                            min_font_size = 10
                            ).generate(words) 
            
            # plot the WordCloud image                        
            plt.figure(figsize = (8, 8), facecolor = None) 
            plt.imshow(wordcloud) 
            plt.axis("off") 
            plt.tight_layout(pad = 0)
        else:
            print(self._add_candidate_info)  


from .charts import *
"""
Dictionary of visualization charts 
"""
chartdict = {
    'imagegrid': ImageGrid,
    'timeline': Timeline,
    'graph': Graph,
    'dimensions': Dimensions,
    'map': Map,
    'tree': Tree,
    'wordcloud': WordCloud,
    'linechart': LineChart,
    'barchart': BarChart,
    'histogram': Histogram,
    'densityplot': DensityPlot,
    'treemap': TreeMap,
    'sunburstchart': SunBurstChart,
    'heatmap': HeatMap,
    'piechart': PieChart,
    'donutchart': DonutChart,
    'boxplot': BoxPlot,
    'violinplot': ViolinPlot,
    'areachart': AreaChart,
    'stackedareachart': StackedAreaChart,
    'scatterchart': ScatterChart,
    'bubblechart': BubbleChart,
    'table': Table
}


import codecs
import os
import re
from setuptools import find_packages, setup

def local_file(file):
  return codecs.open(
    os.path.join(os.path.dirname(__file__), file), 'r', 'utf-8'
)

install_reqs = [
  line.strip()
  for line in local_file('requirements.txt').readlines()
  if line.strip() != ''
]

setup(
    name='myKGlib',
    packages=['myKGlib', 'myKGlib.charts', 'myKGlib.utils'],
    version='0.0.0.8',
    description='VizKG',
    author='Hana',
    install_requires=install_reqs,
    license='MIT',
)


import re
import pandas as pd
import matplotlib.pyplot as plt
import statistics

class Chart:
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Chart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
            kwargs: Arbitrary keyword arguments.
        """
        self.dataframe = dataframe
        self.mode_keyword = self.__set_mode(kwargs.get('mode_keyword'))
        self.figsize = self.__set_figsize(kwargs.get('figsize'))
        
        self._uri_column = self._set_uri_column()
        self._date_column = self._set_date_column()
        self._numerical_column = self._set_numerical_column()
        self._label_column = self._set_label_column()
        self.candidate_viz = self.candidate_form()

    def _set_label_column(self):
        """
        Get object or label column name of dataframe based on regex and object data type
        """
        #Regex pattern
        pattern_url = r"^(?:http(s)?:\/\/)[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$(?<!.[jpg|gif|png|JPG|PNG])" 
        label_column = []

        for i in range (len(self.dataframe.columns)):
            column_name = self.dataframe.columns[i]
            column = self.dataframe[column_name]
            is_uri_column = self.__check_data_per_column(column, pattern_url)
            try:
                if not column_name.startswith(tuple(['coordinate', 'picture'])) and is_uri_column == False and column.dtypes == 'string':
                    label_column.append(column_name)
            except TypeError:
                pass

        return label_column

    def _set_date_column(self):
        """
        Get date column name of dataframe based on date data type
        """
        date_column = [name for name in self.dataframe.columns if self.dataframe[name].dtypes == 'datetime64[ns]']

        return date_column

    def _set_numerical_column(self):
        """
        Get date column name of dataframe based on date data type
        """
        numerical_column = [name for name in self.dataframe.columns if self.dataframe[name].dtypes == 'float64']

        return numerical_column 

    def _set_uri_column(self):
        """
        Get date column name of dataframe based on date data type
        """
        #Regex pattern
        pattern_url = r"^(?:http(s)?:\/\/)[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$(?<!.[jpg|gif|png|JPG|PNG])" 
        uri_column = []

        for i in range (len(self.dataframe.columns)):
            column_name = self.dataframe.columns[i]
            column = self.dataframe[self.dataframe.columns[i]]
            is_uri_column = self.__check_data_per_column(column, pattern_url)
            if is_uri_column:
                uri_column.append(column_name)

        return uri_column

    def candidate_form(self):
        """
        Find candidate form for visualization

        Parameter:
            dataframe (pandas.Dataframe): The data table

        Returns:
            candidate_visualization (list): List of candidate visualization
        """

        candidate_visualization = []
        dimension_column = [name for name in self.dataframe.columns if not name.startswith(tuple(['picture', 'coordinate']))]

        #Add to candidate visualization
        if 'picture' in self.dataframe.columns:
            candidate_visualization.append('ImageGrid')
        if 'coordinate' in self.dataframe.columns:
            candidate_visualization.append('Map')
        if len(self._date_column) >= 1 and len(self._label_column) >= 1:
            candidate_visualization.append('Timeline')
        if len(self._label_column) >= 2 and len(self._uri_column) >= 2:
            candidate_visualization.append('Graph')
            candidate_visualization.append('Tree')
        if len(dimension_column) > 2 :
            candidate_visualization.append('Dimensions')
        if len(self._numerical_column) >= 3:
            candidate_visualization.append('HeatMap')
        if len(self._label_column) >= 1:
            candidate_visualization.append('WordCloud')
        if len(self._label_column) >= 1 and len(self._date_column) >= 1 and len(self._numerical_column) >= 1:
            candidate_visualization.append('AreaChart')
        if len(self._date_column) >= 1 and len(self._numerical_column) >= 2:
            candidate_visualization.append('StackedAreaChart')
        if len(self._numerical_column) >= 2:
            candidate_visualization.append('ScatterChart')
        if len(self._date_column) >= 1 and len(self._numerical_column) >= 1:
            candidate_visualization.append('LineChart')
        if len(self._label_column) <= 2 and len(self._numerical_column) == 1:
            candidate_visualization.append('BarChart')
        if len(self._label_column) >= 2 and len(self._numerical_column) >= 1:
            candidate_visualization.append('TreeMap')
            candidate_visualization.append('SunBurstChart')
        if len(self._numerical_column) >= 1:
            candidate_visualization.append('Histogram')
            candidate_visualization.append('DensityPlot')
        if len(self._label_column) >= 1 and len(self._numerical_column) >= 1:
            candidate_visualization.append('PieChart')
            candidate_visualization.append('DonutChart')
            candidate_visualization.append('BoxPlot')
            candidate_visualization.append('ViolinPlot')
            candidate_visualization.append('BubbleChart')
            candidate_visualization.append('TreeMap')
            candidate_visualization.append('SunBurstChart')
        else:
            candidate_visualization.append('Table')

        return set(candidate_visualization)

    def _is_label_column_exist(self, request=1):
        """
        Check if list exist return True if exist

        Parameters:
            (int) request:number of required column
        
        Returns:
            (boolena) is_exist: True if list exist
        """
        is_exist = False
        if len(self._label_column) >= request:
            is_exist = True
        else:
            miss = request - len(self._label_column)
            print(f"Missing {str(miss)} required label column, instead use one of this available chart: {self.candidate_viz}")

        return is_exist

    def _is_date_column_exist(self, request=1):
        """
        Check if list exist return True if exist

        Parameters:
            (int) request:number of required column
        
        Returns:
            (boolena) is_exist: True if list exist
        """
        is_exist = False
        if len(self._date_column) >= request:
            is_exist = True
        else:
            miss = request - len(self._date_column)
            print(f"Missing {str(miss)} required date column, instead use one of this available chart: {self.candidate_viz}")
        
        return is_exist
    
    def _is_numerical_column_exist(self, request=1):
        """
        Check if list exist return True if exist

        Parameters:
            (int) request:number of required column
        
        Returns:
            (boolena) is_exist: True if list exist
        """
        is_exist = False
        candidate_form = self.candidate_viz
        if len(self._numerical_column) >= request:
            is_exist = True
        else:
            miss = request - len(self._numerical_column)
            print(f"Missing {str(miss)} required numerical column, instead use one of this available chart: {self.candidate_viz}")
        
        return is_exist

    def _is_uri_column_exist(self, request=1):
        """
        Check if list exist return True if exist

        Parameters:
            (int) request:number of required column
        
        Returns:
            (boolena) is_exist: True if list exist
        """
        is_exist = False
        candidate_form = self.candidate_viz
        if len(self._uri_column) >= request:
            is_exist = True
        else:
            miss = request - len(self._uri_column)
            print(f"Missing {str(miss)} required uri column as identifiers, instead use one of this available chart: {self.candidate_viz}")
        
        return is_exist

    def _add_candidate_info(self):
        """
        add information of candidate visualization
        """
        candidate_text = "Instead use of this available chart" + self.candidate_viz

        return candidate_text

    def _check_labels(self):
        """
        Check the requirements for label

        Returns:
            (string) axis_label: label for axis
            (string) group_label: label for group_label
            (string) make_axis_label: label for make group_label
        """
        axis_label = None
        group_label = None
        make_axis_label = None

        if len(self._label_column) > 1:
            unique_dict = {name:len(self.dataframe[name].unique()) for name in (self._label_column)}
            sort_dict = {k: v for k, v in sorted(unique_dict.items(), key=lambda item: item[1])}
            fltr_dict = {name:value for name, value in sort_dict.items() if value <= (len(self.dataframe))}
            key_fltr_dict = list(fltr_dict.keys())
            if len(fltr_dict) < 3:
                axis_label = key_fltr_dict[1]
                make_axis_label = key_fltr_dict[0]
            elif len(fltr_dict) >= 3:
                axis_label = key_fltr_dict[2]
                make_axis_label = key_fltr_dict[1]
                group_label = key_fltr_dict[0]
            else:
                axis_label = key_fltr_dict[0]
        else:
            axis_label = self._label_column[0]
        
        if group_label is not None:
            return axis_label,group_label,make_axis_label
        else:
            return axis_label,make_axis_label

    def _check_orientation(self, axis_label, group_label=None, max_number=6):
        """
        Check the requirements for orientation, returns None if horizontal

        Returns:
            (string) orientation: label for axis
        """
        orientation = None
        num_box = 0
        num_axis = len(self.dataframe[axis_label].unique())
        num_box = 0

        if group_label is not None:
            num_group = len(self.dataframe[group_label].unique())
            num_box = num_axis + num_group
        else:
            num_box = num_axis

        if num_box > max_number:
            orientation = 'Horizontal'

        return orientation

    def _check_numerical_columns(self):
        """
        Check the requirements for numerical label

        Returns:
            (string) values_label: label for values
            (string) hover_label: label for hover text
        """
        values_label = None
        hover_label = None

        min_dict = {name:statistics.median(list(self.dataframe[name])) for name in (self._numerical_column)}
        sort_dict = {k: v for k, v in sorted(min_dict.items(), key=lambda item: item[1])}
        if len(sort_dict) > 1:
            values_label = list(sort_dict.keys())[1]
            hover_label = list(sort_dict.keys())[0]
        else:
            values_label = list(sort_dict.keys())[0]
        
        return values_label,hover_label

    def __check_data_per_column(self,column, pattern):
        """
        Check entire data per column of dataframe if matched with regex pattern

        Parameters:
            (pandas.Dataframe) column: column name of dataframe
            (string) pattern: regex pattern

        Returns:
            (boolen) boolean_check: The result table             
        """
        boolean_check = False
        for datapoint in range(len(column)):
            data = column.iloc[datapoint]
            try:
                if re.match(pattern, data):
                    boolean_check = True
            except TypeError:
                pass
                
        return boolean_check

    @staticmethod
    def __set_mode(mode_input):
        """
        Setter of mode based on mode input

        Parameters:
            (bool) mode_input: The mode input

        Returns:
            (bool) mode: The result mode  
        """
        mode = None
        if mode_input is not None and isinstance(mode_input, bool):
            mode = mode_input
        else:
            mode = None
        
        return mode

    @staticmethod
    def __set_figsize(figsize_input):
        """
        Setter of figsize based on figsize input

        Parameters:
            (tuple) figsize_input: The figsize input

        Returns:
            (tuple) figsize: The result figsize  
        """
        figsize = None
        is_numeric_value = None

        if figsize_input is not None and len(figsize_input) == 2:
            is_numeric_value = all(isinstance(v, int) or isinstance(v, float) for v in figsize_input)
        else:
            is_numeric_value = False
            
        if is_numeric_value:
            figsize = figsize_input
        else:
            figsize = None

        return figsize


import re
import pandas as pd
from pandas import json_normalize
from SPARQLWrapper import SPARQLWrapper
from difflib import SequenceMatcher

from myKGlib.chartdict import chartdict

def set_chart(chart_input):
    """
    Setter of chart based on chart input

    Parameters:
        (string) chart_input: The chart input

    Returns:
        (list) chart: The available chart   
    """
    chart = None
    charts = chartdict.keys()

    if chart_input is not None:
        lowercase_input = chart_input.lower()
        
        highest_prob = 0
        for name in charts:
            prob_now = SequenceMatcher(None, lowercase_input, name).ratio()
            if prob_now > highest_prob and prob_now >= 0.5:
                highest_prob = prob_now
                chart = name

        if lowercase_input in charts:
            chart = lowercase_input

        if chart is None:
            raise Exception("No available chart")

    else:
        chart = None

    return chart


def set_dataframe(sparql_query, sparql_endpoint):
    """
    Query the endpoint with the given query string and format the result table

    Parameters:
        (string) sparql_query: The sparql query.
        (string) sparql_endpoint: The sparql endpoint

    Returns:
        (pandas.Dataframe) result_table: The table of result    
    """

    sparql = SPARQLWrapper(sparql_endpoint)  

    sparql.setQuery(sparql_query)
    sparql.setReturnFormat('json')

    results = sparql.query().convert()
    table  = json_normalize(results["results"]["bindings"])

    data_table = table[[column_name for column_name in table.columns if column_name.endswith('.value')]]
    data_table.columns = data_table.columns.str.replace('.value$', '', regex=True)
    rename_column_table = __rename_column_table(data_table)
    change_dtype_table = __change_dtypes(rename_column_table)
    result_table = change_dtype_table
    
    return result_table


def __rename_column_table(dataframe):
    """
    Rename column of dataframe based on regex validity check

    Parameters:
        (pandas.Dataframe) dataframe: The table

    Returns:
        (pandas.Dataframe) dataframe: The result table             
    """

    #Regex pattern
    pattern_url = r"^(?:http(s)?:\/\/)[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$(?<!.[jpg|gif|png|JPG|PNG])" 
    pattern_img = r"^http(s)?://(?:[a-z0-9\-]+\.)+[a-z]{2,6}(?:/[^/#?]+)+\.(?:jpg|jpeg|gif|png|JPG|JPEG|Jpeg)$"        
    pattern_coordinate = r"^Point"

    for i in range (len(dataframe.columns)):
        column = dataframe[dataframe.columns[i]]
        is_picture_column = __check_data_per_column(column, pattern_img)
        is_coordinate_column = __check_data_per_column(column, pattern_coordinate)
        if is_picture_column:
            dataframe =  dataframe.rename(columns={dataframe.columns[i]: "picture"}, errors="raise")
        elif is_coordinate_column:
            dataframe =  dataframe.rename(columns={dataframe.columns[i]: "coordinate"}, errors="raise")
        else:
            pass

    return dataframe

def __check_data_per_column(column, pattern):
    """
    Check entire data per column of dataframe if matched with regex pattern

    Parameters:
        (pandas.Dataframe) column: column name of dataframe
        (string) pattern: regex pattern

    Returns:
        (boolen) boolean_check: The result table             
    """
    boolean_check = False
    for datapoint in range(len(column)):
        data = column.iloc[datapoint]
        try:
            if re.match(pattern, data):
                boolean_check = True
        except TypeError:
            pass
            
    return boolean_check

def __change_dtypes(dataframe):
    """
    Change data type column of dataframe

    Parameters:
        (pandas.Dataframe) dataframe: The table

    Returns:
        (pandas.Dataframe) table: The result table             
    """

    for column in dataframe:
        try:
            dataframe[column] = dataframe[column].astype('string')
        except ValueError:
            pass

    for column in dataframe:
        try:
            dataframe[column] = dataframe[column].astype('datetime64')
        except ValueError:
            pass

    for column in dataframe:
        try:
            dataframe[column] = dataframe[column].astype('float64')
        except (ValueError, TypeError):
            pass

    return dataframe


from .chart import Chart
import plotly.express as px

class SunBurstChart(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the SunBurstChart object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating SunBurstChart visualization

        Returns:
            (list) label_column: label name
            (list) numerical_column: list of numerical column
        """
        label_column = None
        numerical_column = None
        
        if self._is_label_column_exist(1):
            label_column = self._label_column
            if self._is_numerical_column_exist(1):
                numerical_column = self._numerical_column

        
        return label_column, numerical_column    

    def plot(self):
        """
        Generate SunBurstChart visualization
        """
        label_column, numerical_column  = self._check_requirements()

        if label_column is not None and numerical_column is not None:
            values_label,hover_label = self._check_numerical_columns()
            if hover_label is not None:
                #plot
                fig = px.sunburst(self.dataframe, values=values_label, path=label_column, hover_data=[hover_label], labels={hover_label:hover_label})
                fig.show()
            else:
                fig = px.sunburst(self.dataframe, values=values_label, path=label_column)
                fig.show()                






from .chart import Chart
import networkx as nx
import plotly.graph_objects as go
import matplotlib.pyplot as plt

class Graph(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Graph object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
            kwargs (dictionary): Arbitrary keyword arguments.
        """
        Chart.__init__(self, dataframe, kwargs)

    def _check_requirements(self):
        """
        Check the requirements for generating graph visualization

        Returns:
            (list) filter_column: list of filter label name
            (bool) is_label_column: if column not uri column
        """
        filter_column = None
        is_label_column = False

        if self._is_uri_column_exist(2):
            if len(self._uri_column) > len(self._label_column):
                filter_column = self._uri_column
            else:
                filter_column = self._label_column
                is_label_column = True
        
        return filter_column, is_label_column

    def _check_variable(self):
        """
        Check the required var for generating graph visualization

        Returns:
            (list) filter_column: list of filter label name        
        """
        filter_column = None
        required_var = ['source_node', 'target_node', 'edge_label']
        exist_var = [name for name in self.dataframe.columns if name.startswith(tuple(required_var)) and self.dataframe[name].dtypes == 'string']
        miss_var = list(set(required_var)-set(exist_var))

        if len(miss_var) > 0:
            raise Exception(f"Missing required variable: {miss_var}")
        else:
            filter_column = required_var

        return filter_column
        

    def plot(self):
        """
        Generate graph visualization
        """
        is_label_column = True
        filter_column = None

        #check mode keyword
        if self.mode_keyword is not None:
            filter_column = self._check_variable()
        else:
            filter_column, is_label_column = self._check_requirements()

        #plot
        if filter_column is not None:
            #check if param figsize exist
            if self.figsize is not None:
                plt.figure(figsize=self.figsize)
            else:
                plt.figure(figsize=(20,15))
            try:
                #check if edge label exist
                if len(filter_column) > 2:
                    #check if label exist (not uri)
                    if is_label_column:
                        graph, positions, edge_labels = self.create_graph_nx(filter_column[0], filter_column[1], filter_column[2])
                    else:
                        graph, positions, edge_labels = self.create_graph_nx(filter_column[0], filter_column[2], filter_column[1])
                    nx.draw_networkx(graph, positions, arrowsize=15)
                    nx.draw_networkx_edge_labels(graph, pos=positions, edge_labels=edge_labels, font_color='r')
                else:
                    graph, positions, edge_labels = self.create_graph_nx(filter_column[0], filter_column[1])
                    nx.draw_networkx(graph, positions, arrowsize=15)
            finally:
                plt.show()

    def create_graph_nx(self, source_column, target_column, edge_column=None):
        """
        Create graph networkx

        Paramaters:
            (list) node_list: list of node
            (list) filter_column: list of parent and child name column

        Returns:
            (networkx.DiGraph) Graph: Digraph graph
        """
        Graph = nx.DiGraph()
        
        #add edges and edge_label to graph
        edge_label = {}
        for key, node in self.dataframe.iterrows():
            Graph.add_edges_from([(node[source_column],node[target_column])])
            if edge_column is not None:
                edge_label[(node[source_column],node[target_column])] = node[edge_column]


        #Getting positions for each node.
        positions = nx.kamada_kawai_layout(Graph)

        return Graph, positions, edge_label

    def plot_(self):
        """
        Generate graph visualization
        """
        #filter_column
        filter_column, is_label_column = self._check_requirements()

        if filter_column is not None:
            node_list = self.create_node_list(filter_column)
            graph, positions = self.create_graph_nx(filter_column, node_list)
            edge_trace = self.create_edge_trace(graph, filter_column[-1])
            node_trace = self.create_node_trace(graph)

            annotations = self.add_annotations(graph, positions)

            figure = self.create_figure_go(edge_trace, node_trace, annotations)
            figure.show()
            print(f"{nx.info(graph)}\nDensity graph: {nx.density(graph)}")

    def create_node_list(self, filter_column):
        """
        Create the node list

        Paramaters:
            (list) filter_column: list of source and target name column

        Returns:
            (list) node_list: list of node
        """
        source = list(self.dataframe[filter_column[0]].unique())
        target = list(self.dataframe[filter_column[1]].unique())
        node_list = set (source + target)

        return node_list

    def create_graph_nx_(self, filter_column, node_list):
        """
        Create graph networkx

        Paramaters:
            (list) node_list: list of node
            (list) filter_column: list of source and target name column

        Returns:
            (networkx.DiGraph) Graph: Digraph graph
        """
        Graph = nx.DiGraph()

        #add node to graph
        for node in node_list:
            Graph.add_node(node)
        
        #add edges to graph
        for key, node in self.dataframe.iterrows():
            Graph.add_edges_from([(node[filter_column[0]],node[filter_column[1]])])

        #Getting positions for each node.
        positions = nx.spring_layout(Graph, k=2)

        #Adding positions of the nodes to the graph
        for node, position in positions.items():
            Graph.nodes[node]['position'] = position

        return Graph, positions

    def create_edge_trace(self, Graph, edge_column):
        """
        Create edge_trace label

        Paramaters:
            (networkx.DiGraph) Graph: Digraph graph        

        Returns:
            (plotly.Scatter) edge_trace: edge_trace label
        """
        edge_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            line=dict(width=0.5,color='#888'),
            hoverinfo='text',
            mode='lines')

        #Add edges as disconnected lines in a single trace 
        for edge in Graph.edges():
            x0, y0 = Graph.nodes[edge[0]]['position']
            x1, y1 = Graph.nodes[edge[1]]['position']
            edge_trace['x'] += tuple([x0, x1, None])
            edge_trace['y'] += tuple([y0, y1, None])
            
        for value in self.dataframe[edge_column]:
            edge_info = '<' + str(value)
            edge_trace['text']+=tuple([edge_info])

        return edge_trace

    def create_node_trace(self, Graph):
        """
        Create node_trace object

        Paramaters:
            (networkx.DiGraph) Graph: Digraph graph     

        Returns:
            (plotly.Scatter) node_trace: node_trace object
        """
        node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode='markers',
            hoverinfo='text',
            marker=dict(
                showscale=True,
                colorscale='Viridis',
                reversescale=True,
                color=[],
                size=12,
                colorbar=dict(
                    thickness=15,
                    title='Node Connections',
                    xanchor='left',
                    titleside='right'
                ),
                line=dict(width=2)))
        
        #Add nodes as a scatter trace
        for node in Graph.nodes():
            x, y = Graph.nodes[node]['position']
            node_trace['x'] += tuple([x])
            node_trace['y'] += tuple([y])

        #Add hover text
        node_trace = self.add_hover_text(Graph, node_trace) 

        return node_trace

    def add_hover_text(self, Graph, node_trace):
        """
        Creating the hover text, that will display the info of node on the figure

        Paramaters:
            (networkx.DiGraph) Graph: Digraph graph 
            (plotly.Scatter) node_trace: node_trace object    

        Returns:
             (plotly.Scatter) node_trace: node_trace object with hover text

        """
        #Create degree dictionary (degree, in_degree, out_degree)
        degree_dict = {}
        for key, value in (Graph.degree()):
            degree_dict.setdefault(key, []).append(value)
        for key, value in (Graph.in_degree()):
            degree_dict.setdefault(key, []).append(value)
        for key, value in (Graph.out_degree()):
            degree_dict.setdefault(key, []).append(value)

        for key, value in enumerate(degree_dict):
            node_trace['marker']['color']+=tuple([(degree_dict[value][0])])
            node_info = str(value) + ', Degree:'+ str(degree_dict[value][0]) + " In:" + str(degree_dict[value][1]) + " Out:" + str(degree_dict[value][2])
            node_trace['text']+=tuple([node_info])
        
        return node_trace

    def add_annotations(self, Graph, positions):
        """
        Creating the annotations, that will display the node name on the figure

        Paramaters:
            (networkx.DiGraph) Graph: Digraph graph
            (dict) positions: dictionary of positions for each node in graph    

        Returns:
            (list) annotations: list of annotations        
        """
        # Creating the annotations, that will display the node name on the figure
        annotations = []
        for node, adjacencies in enumerate(Graph.adjacency()):
            annotations.append(
                dict(x=positions[adjacencies[0]][0],
                    y=positions[adjacencies[0]][1],
                    text=adjacencies[0], # node name that will be displayed
                    xanchor='left',
                    xshift=10,
                    font=dict(color='black', size=11),
                    showarrow=False, arrowhead=1, ax=-10, ay=-10)
                )
        
        return annotations


    def create_figure_go(self, edge_trace, node_trace, annotations):
        """
        Create figure of graph

        Paramaters:
            (plotly.Scatter) node_trace: node_trace object
            (plotly.Scatter) edge_trace: edge_trace object                  

        Returns:
            (list) annotations: list of annotations  
        """
        #Coloring based on the number of connections of each node.
        figure = go.Figure(data=[edge_trace, node_trace],
                    layout=go.Layout(
                        titlefont=dict(size=16),
                        showlegend=False,
                        hovermode='closest', 
                        margin=dict(b=20,l=5,r=5,t=40),
                        annotations=annotations, 
                        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False)))
        return figure

    def _add_info_graph(self, Graph):
        """
        Add infomation of graph

        Paramaters:
            (networkx.DiGraph) Graph: Digraph graph    

        Returns:
            (string) number_of_nodes: number of nodes in graph
            (string) number_of_edges: number of edges in graph
            (string) info_density: information of density graph
        """
        number_of_nodes = Graph.number_of_nodes()
        number_of_edges = Graph.number_of_edges()
        info_density = nx.density(Graph)

        return number_of_nodes, number_of_edges, info_density  




from .chart import Chart
import folium
from IPython.display import display

class Map(Chart):
    def __init__(self, dataframe, kwargs):
        """
        Constructs all the necessary attributes for the Map object

        Parameters:
            dataframe (pandas.Dataframe): The dataframe
        """
        Chart.__init__(self, dataframe, kwargs)
    
    def _check_requirements(self):
        """
        Check the requirements for generating tree visualization

        Returns:
            (list) popup_data: list of label name
        """
        popup_data = None
        if hasattr(self.dataframe, 'coordinate'):
            new_data = self._add_point()
            if len(self._label_column) == 0:
                popup_data = new_data.coordinate_point
            else:
                popup_data = new_data[self._label_column[0]]
        else:
            print(f"Data has no attribute 'coordinate'")
            print(self._add_candidate_info)
        
        return popup_data

    def _add_point(self):
        """
        Add coordinate column for coordinate folium map

        Returns:
            (pandas.Dataframe): Dataframe with new coordinate column
        """
        data = self.dataframe.copy()    
        #Get coordinate data (latitude and longitude)
        data['coordinate_point'] = data['coordinate']
        dataframe_new = data.apply(lambda S:S.str.strip('Point()'))
        new = dataframe_new[dataframe_new.columns[-1]].str.split(" ", n = 1, expand = True)
        new = new.astype('float64')
        data['coordinate'] = new.apply(lambda x: list([x[1], x[0]]),axis=1)

        return data

    def plot(self):
        """
        Generate map visualization
        """
        popup_data = self._check_requirements()

        if popup_data is not None:
            new_data = self._add_point()
            #Initiate map folium object
            maps = folium.Map()

            #Marked the map folium object
            for i in range (len(new_data)):
                folium.Marker(
                    location=new_data.coordinate[i],
                    popup=popup_data[i]
                ).add_to(maps)

            display(maps)                